;---------------------------------------------------------------------------------------------------
; name:			geoms_online_conversion.pro
; description: 	Wrapper for EVDC GEOMS on-line conversion tool
;
; Note: The output directory of the GEOMS compliant file(s) and log file are currently hard-wired
; into the call to idlcr8hdf (/M is the same directory as the Metadata file). This can be changed
; as required.
;
; 2018-11-16, v0.1	Ian Boyd initial implementation
; 2020-10-20, v1.1  Ian Boyd compile with idlcr8hdf.pro v4.0b56
; 2020-10-26, v1.2  Ian Boyd compile with idlcr8hdf.pro v4.0b57
; 2020-12-11, v1.3  Ian Boyd compile with idlcr8hdf.pro v4.0b58
; 2021-06-17, v1.4  Ian Boyd compile with idlcr8hdf.pro v4.0b59
; 2022-08-09, v1.5  Ian Boyd compile with idlcr8hdf.pro v4.0b60
; 2023-12-18, v1.6  Ian Boyd compile with idlcr8hdf.pro v4.0b61
; 2024-01-12, v1.7  Ian Boyd compile with idlcr8hdf.pro v4.0b62
; 2024-09-12, v1.8  Ian Boyd compile with idlcr8hdf.pro v4.0b64
; 2024-10-29, v1.9  Ian Boyd compile with idlcr8hdf.pro v4.0b65
; 2024-11-26, v1.10 Ian Boyd Add run-time error handling capability and compile with 
;                   idlcr8hdf.pro v4.0b66 and idlcr8ascii.pro v4.0b28
; 2024-12-18, v1.11 Ian Boyd compile with idlcr8hdf.pro v4.0b67 
; 2025-01-24, v1.12 Ian Boyd compile with idlcr8hdf.pro v4.0b68                 
;---------------------------------------------------------------------------------------------------

PRO geoms_online_conversion

args = command_line_args()
iNrArgs = size(args, /N_ELEMENTS)

if ( iNrArgs ne 4 ) then begin
  print
  print, 'GEOMS online conversion'
  print, '  Release v1.12, 2025-01-24'
  print, '  Does conversion of input data and metadata files to GEOMS compliant HDF4/5 and netCDF files.'
  print
  print, 'usage: idl -rt=geoms_online_conversion.sav -args METAFILE ORIGDATAFILE(s) TAVFILE GEOMSFORMAT'
  print
  print, '1: METAFILE is the path and file name of the Metadata template file'
  print, '2: ORIGDATAFILE(s) is the path and file name of the ASCII data file, or a path and filespec'
  print, '   of a set of ASCII data files e.g. /data_dir/input.data or /data_dir/input*.data'
  print, '3: TAVFILE is the path of the TAV file, you want to validate against'
  print, '4. GEOMSFORMAT is the output format of the GEOMS file (H4, H5, NC)
  print, '------------------------------------------------------------------------------------------------'
  print

  QA_error_code = 3
  exit, status=QA_error_code
end

chFileMeta = STRTRIM( args( 0 ), 2 )
chFileData = STRTRIM( args( 1 ), 2 )
chFileTAV = STRTRIM( args( 2 ), 2 )
chFileFormat = STRTRIM( STRUPCASE( args( 3 ) ), 2 )

QA_error_code = 3

if ( file_test( chFileTAV ) eq 0 ) then begin
  QA_error_code = 3
  print, 'ERROR: TAV file not found.'
endif else QA_error_code = 0

if ( file_test( chFileMeta ) eq 0 ) then begin
  QA_error_code = 3
  print, 'ERROR: Metadata Template file not found.'
endif else QA_error_code = 0

arFileData=File_Search(chFileData, /Fully_Qualify_Path, count = fcnt )
if fcnt eq 0 then begin
  QA_error_code = 3
  print, 'ERROR: Data file(s) not found.'
endif else QA_error_code = 0

AllowedFormats=['H4','H5','NC']
gi = WHERE(chFileFormat eq AllowedFormats, gcnt)
if gcnt eq 0 then begin
  QA_error_code = 3
  print, 'ERROR: Output format must be H4, H5 or NC.'
endif else QA_error_code = 0


if ( QA_error_code eq 0 ) then begin

  CATCH, error_status
  IF error_status NE 0 THEN BEGIN
    ;Run-time error so exit nicely
    CLOSE,/ALL
    ;Delete idlcr8ascii.log and idlcr8log.log if they exist
    logdir=FILE_DIRNAME(chFileMeta,/MARK_DIRECTORY)
    FILE_DELETE,logdir+'idlcr8ascii.log',/QUIET
    ;FILE_DELETE,logdir+'idlcr8log.log',/QUIET
    ;Write Run-time error to the log file
    logfile=logdir+'idlcr8log.log'
    etxt=' Please contact Ian Boyd at iboyd@bryanscientific.org and nadirteam@nilu.no'
    OPENW,lu,logfile, /GET_LUN, /Append
    PRINTF,lu,''
    PRINTF,lu,'  RUN-TIME ERROR: '+!ERROR_STATE.MSG
    PRINTF,lu,''
    PRINTF,lu,'99 (Failed with Run-Time error.'+etxt+')'
    FREE_LUN,lu
    CATCH,/CANCEL
    QA_error_code=99
    exit, status=QA_error_code
  ENDIF

  case 1 of
    chFileFormat eq 'H4': idlcr8hdf, chFileMeta, arFileData, chFileTav, 'M', /AVK, /Log
    chFileFormat eq 'H5': idlcr8hdf, chFileMeta, arFileData, chFileTav, 'M', /H5, /AVK, /Log
    chFileFormat eq 'NC': idlcr8hdf, chFileMeta, arFileData, chFileTav, 'M', /NC, /AVK, /Log
  endcase

end

exit, status=QA_error_code

END
